import tkinter as tk
from typing import Any, Self
import customtkinter as ctk
from flowmatic import gui
from flowmatic.gui.elements.element_infos import Element


class Display:
    master: tk.Frame
    frame: tk.Frame
    label: str
    __value: None = None
    elements: list[Element]

    @property
    def value(self) -> None:
        return self.__value

    @value.setter
    def value(self, value) -> None:
        pass

    def __init__(self, label: str, elements: list[Element] | None = None) -> None:
        self.label = label
        self.elements = elements or []

    def __call__(
        self,
        master: tk.Frame,
    ) -> Self:
        self.master = master
        self.frame = ctk.CTkFrame(
            self.master,
            # fg_color=ctk.ThemeManager.theme["CTkFrame"]["fg_color"],
        )
        return self

    def build(self, **kwargs: dict[str, Any]) -> Self:
        self.frame.pack()
        ctk.CTkLabel(self.frame, text=self.label).pack(**gui.pack_defaults)
        for element in self.elements:
            gui.screens.ElementScreen.build_element(element, self.frame)
        return self

    def pack(self, element: tk.Widget, **kwargs) -> None:
        element.pack(**kwargs)



import customtkinter as ctk

from flowmatic import gui
from flowmatic.gui.screens.screen import Screen
from ...gui.elements import Button


class TitleScreen(Screen):
    """Title screen.

    Args:
        master (tk.Tk): Master of screen."""

    def build(  # pylint: disable=arguments-differ
        self, title: str, button: Button
    ) -> None:
        """Build screen.

        Args:
            title (str): Title of screen.
            button (gui.ButtonInfo): Button to show.
        """
        ctk.CTkLabel(self.master, text=title).pack(**gui.pack_defaults)
        ctk.CTkButton(self.master, text=button[0], command=button[1]).pack(
            **gui.pack_defaults,
        )



from typing import Tuple
import customtkinter as ctk
import tkinterdnd2
import tkinter.filedialog as tkfd
from CTkMessagebox import CTkMessagebox


from flowmatic.gui.gui import GUI  # pylint: disable=import-error

from .screens.screen import Screen
from ..flows.flow import Flow


class CustomTk(ctk.CTk):
    def __init__(self, fg_color: str | Tuple[str, str] | None = None, **kwargs):
        ctk.CTk.__init__(self, fg_color=fg_color, **kwargs)
        self.TkdndVersion = tkinterdnd2.TkinterDnD._require(self)


class TKGUI(GUI):
    root: CustomTk
    title: str
    start_screen: Screen
    geometry: str
    light_mode: bool

    def __init__(
        self,
        title: str,
        start_screen: type[Screen],
        geometry: str = "1280x720",
        light_mode: bool = False,
    ) -> None:
        self.root = CustomTk()
        self.title = title
        self.start_screen = start_screen()
        self.geometry = geometry
        self.light_mode = light_mode

    def start(self, start_screen: Screen | None) -> None:
        """Start GUI.
        Args:
            screen (type[Screen]): Screen to start with."""
        self.setup()
        self.build_screen(start_screen or self.start_screen)
        self.root.mainloop()

    def setup(self) -> None:
        ctk.set_appearance_mode("light" if self.light_mode else "dark")
        self.root.geometry(self.geometry)
        self.root.title(self.title)
        self.root.resizable(False, False)
        self.root.focus_set()
        self.root.configure(bg=ctk.ThemeManager.theme["CTk"]["fg_color"])

    def file_dialog(
        self, title: str, max_files: int, filetypes: list[tuple[str, str]]
    ) -> str | tuple[str, ...]:
        """Open file dialog.
        Args:
            title (str): Title of the dialog.
            max_files (int): Maximum number of files to select.
            filetypes (list[tuple[str, str]]): Filetypes to show.
        Returns:
            str | None: Path to file or None if no file selected."""
        if max_files == 1:
            return tkfd.askopenfilename(title=title, filetypes=filetypes)
        return tkfd.askopenfilenames(title=title, filetypes=filetypes)

    def start_flow(self, flow: Flow) -> None:
        self.clear_screen()
        self.current_flow = flow
        flow.start()

    def switch_screen(self, screen: Screen) -> None:
        """Switch to a new screen.

        Args:
            screen (type[Screen]): Screen to show."""
        self.clear_screen()
        self.build_screen(screen)

    def build_screen(self, screen: Screen) -> None:
        """Build screen.

        Args:
            screen (type[Screen]): Screen to build."""
        screen(self.root).build()
        self.current_screen = screen

    def clear_screen(self) -> None:
        for widget in self.root.winfo_children():
            widget.destroy()

    def show_error(self, error: Exception | str) -> None:
        CTkMessagebox(title="Error", message=f"{error}", icon="cancel")

    def update(self) -> None:
        self.root.update()

    def quit(self) -> None:
        self.root.quit()



from abc import ABC
import json
from typing import Any, Self

from flowmatic.util import SavedClassLoadException


class SavedClass(ABC):
    def __init__(self, path: str):
        self.path = path

    def set(self, kwargs: dict[str, Any]) -> Self:
        """Set attributes.

        Args:
            kwargs (dict[str, Any]): Attributes to set.

        Returns:
            Self: Self."""
        for key, value in kwargs.items():
            setattr(self, key, value)
        return self

    def items(self):
        _dict = self.__dict__.copy()
        _dict.pop("path")
        return _dict.items()

    def save(self):
        if self.path:
            with open(self.path, "w", encoding="utf-8") as file:
                json.dump(self.__dict__, fp=file)

    def load_from_disk(self) -> None:
        with open(self.path, "r", encoding="utf-8") as file:
            try:
                json_data = json.load(file)
                for key, value in json_data.items():
                    setattr(self, key, value)
            except Exception as exc:
                raise SavedClassLoadException(self.path) from exc

    def create_file(self) -> None:
        f = open(self.path, "w", encoding="utf-8")
        f.close()

    @classmethod
    def load(cls, path: str) -> Self:
        """Load the class from disk.

        Args:
            path (str): Path to load from.

        Returns:
            Self: The loaded class.

        Raises:
            SavedClassLoadException: If the class could not be loaded."""
        self = cls(path)
        try:
            self.load_from_disk()
        except FileNotFoundError:
            self.create_file()

        return self



from .gui import *
from .flows import *
from .forms import *
from .server import Server
from .util import *
from .flowmatic import App


server = Server()


def run(app: App):
    """Start app.
    Args:
        screen (type[Screen]): Screen to start with."""
    server.app = app
    server.app.start()


def get_app() -> App:
    return server.app


def set_start_screen(screen_t: type[Screen]):
    """Set start screen.
    Args:
        screen (type[Screen]): Screen to start with."""
    screen = screen_t()
    server.app.start_screen = screen
    server.app.gui.start_screen = screen


def show_start_screen():
    """Show start screen."""
    server.app.show_start_screen()


def show_screen(screen: Screen):
    """Show screen.
    Args:
        screen (type[Screen]): Screen to show."""
    server.app.show_screen(screen)


def show_error(error: Exception | str):
    server.app.gui.show_error(error)


def start_flow(flow: Flow):
    """Start flow.
    Args:
        flow (Flow): Flow to start."""
    server.app.start_flow(flow)


def get_settings() -> Settings:
    """Get settings.
    Returns:
        Settings: Settings."""
    return server.app.settings


def set_settings(settings: dict[str, str]) -> None:
    """Set settings.
    Args:
        settings (dict[str, str]): Settings."""
    server.app.settings.set(settings)
    server.app.settings.save()


def file_dialog(
    title: str, max_files: int, filetypes: list[tuple[str, str]]
) -> tuple[str, ...]:
    """Open file dialog.
    Args:
        title (str): Title.
        max_files (int): Max files.
        filetypes (list[tuple[str, str]]): Filetypes.
    Returns:
        str | None: File path."""
    return server.app.file_dialog(title, max_files, filetypes)


def update_gui():
    server.app.gui.update()


def quit_app():
    server.app.quit()



from typing import Callable, Any


class FormValidation:
    field_label: str
    condition: Callable[[Any], bool]
    message: str

    def __init__(
        self, field_label: str, condition: Callable[[Any], bool], message: str
    ) -> None:
        self.field_label = field_label
        self.condition = condition
        self.message = message

    def validate(self, values: dict[str, int | str | list[str]] | None) -> None:
        assert values, "Form is empty."
        value = values.get(self.field_label, None)
        assert value is not None, f"{self.field_label} is empty."
        assert self.condition(value), f"{self.field_label}: {self.message}"



import asyncio
from dataclasses import dataclass
from enum import Enum
from tkinter import ttk
from typing import Any, AsyncGenerator, Callable, Self, overload

import tkinter as tk
import customtkinter as ctk
import flowmatic as fm
from flowmatic import gui

from flowmatic.gui.screens.screen import Screen


class Status(Enum):
    DEFAULT = "default"
    SUCCESS = "green"
    WARNING = "yellow"
    ERROR = "red"
    SPECIAL = "purple"


class Result:
    result: tuple[str, ...]
    status: Status

    def __init__(self, *args: str, status: Status | None = None) -> None:
        self.result = args
        self.color = status or Status.DEFAULT


class ResultScreen(Screen):
    results: list[Result]
    generator: AsyncGenerator[Result, Any] | None = None
    gen_length: int | None = None
    headers: tuple[str, ...]
    on_select: Callable[[tuple[str, ...]], None]
    show_index: bool = True
    topbar: Screen | None = None
    bottombar: Screen | None = None
    __tree: ttk.Treeview
    __progressbar: ctk.CTkProgressBar
    __progress: tk.IntVar

    @property
    def progress(self) -> float:
        if self.gen_length:
            return self.__progress.get() / self.gen_length
        return 1

    @progress.setter
    def progress(self, value: int) -> None:
        self.__progress.set(value)

    def __init__(self) -> None:
        if not hasattr(self, "results"):
            self.results = []
        self.__progress = tk.IntVar(value=0)

    def __call__(self, master: tk.Tk | tk.Frame) -> Self:
        self.master = master  # pylint: disable=attribute-defined-outside-init
        return self

    def build(self) -> None:
        frame = ctk.CTkFrame(self.master)
        frame.pack(
            **gui.pack_defaults,
            fill="both",
        )

        if self.topbar:
            top_frame = ctk.CTkFrame(frame)
            top_frame.pack()
            self.topbar(top_frame).build()  # pylint: disable=not-callable

        inner_frame = ctk.CTkFrame(frame)
        inner_frame.pack(
            **gui.pack_defaults,
            fill="both",
        )
        progress_frame = ctk.CTkFrame(inner_frame)
        self.build_progress_bar(progress_frame)
        if self.gen_length:
            progress_frame.pack()

        if self.bottombar:
            bottom_frame = ctk.CTkFrame(frame)
            bottom_frame.pack()
            self.bottombar(bottom_frame).build()  # pylint: disable=not-callable

        self.__tree = tree = ttk.Treeview(
            inner_frame,
            columns=("index", *self.headers),
            displaycolumns=(
                used_headers := ("index", *self.headers)
                if self.show_index
                else self.headers
            ),
            show="headings",
        )
        tree.bind(
            "<<TreeviewSelect>>", lambda _: self.do_selection(tree.selection()[0])
        )
        tree.tag_configure(Status.ERROR.value, background="red", foreground="black")
        tree.tag_configure(Status.SUCCESS.value, background="green", foreground="black")
        tree.tag_configure(
            Status.WARNING.value, background="yellow", foreground="black"
        )
        tree.tag_configure(
            Status.SPECIAL.value, background="purple", foreground="black"
        )
        tree.pack(fill="both", expand=True)

        for header in used_headers:
            tree.heading(header, text=header.capitalize())

        if self.results:
            for index, result in enumerate(self.results):
                tree.insert(
                    parent="",
                    index="end",
                    values=(index + 1, *result.result),
                    tags=(result.status.value,),
                )

        fm.update_gui()

        if self.generator:
            asyncio.run(self.generate_results())

    async def generate_results(self) -> None:
        if not self.generator:
            return
        async for index, result in aenumerate(
            self.generator, r_len := len(self.results) + 1
        ):
            self.results.append(result)
            self.__tree.insert(
                parent="",
                index="end",
                values=(index, *result.result),
                tags=(result.color.value,),
            )
            self.progress = index - r_len + 1
            self.__progressbar.set(self.progress)
            fm.update_gui()

    def do_selection(self, index_str: str) -> None:
        index = int(index_str[1:], base=16) - 1  # indexes in table start at 1
        self.on_select(self.results[index].result)

    def build_progress_bar(self, frame: ctk.CTkFrame) -> None:
        ctk.CTkLabel(frame, textvariable=self.__progress).pack(
            side="left", padx=(20, 0)
        )
        ctk.CTkLabel(frame, text=f" / {self.gen_length}").pack(side="left")
        self.__progressbar = ctk.CTkProgressBar(master=frame)
        self.__progressbar.pack(padx=20, pady=10, side="right")
        self.__progressbar.set(0)


async def aenumerate(
    asequence: AsyncGenerator[Result, Any], start=0
) -> AsyncGenerator[tuple[int, Result], Any]:
    """Asynchronously enumerate an async iterator from a given start value"""
    index = start
    async for elem in asequence:
        yield index, elem
        index += 1



from typing import Any, Protocol, Self
import tkinter as tk


class Field(Protocol):
    master: tk.Frame
    label: str
    __value: Any

    @property
    def value(self) -> int | str | list[str] | None:
        return self.__value

    @value.setter
    def value(self, value: Any) -> None:
        self.__value = value

    def __init__(
        self,
        label: str,
    ) -> None:
        ...

    def __call__(self, master: tk.Frame) -> Self:
        ...

    def build(self) -> Self:
        ...

    def pack(  # pylint: disable=missing-function-docstring
        self, element: tk.Widget, **kwargs: dict[str, Any]
    ) -> None:
        ...



import os
import platform
import tkinter as tk
from typing import Iterable, Literal, Self

import customtkinter as ctk
from PIL import Image
from tkinterdnd2 import DND_FILES

import flowmatic
from flowmatic import gui
from flowmatic.util import Icons
from flowmatic.util.exceptions import TooManyFilesError

MAX_COLUMNS = 4
IMAGE_WIDTH = 130


class FileField:
    master: tk.Frame
    frame: tk.Frame
    label: str
    __value: list[str]
    __msg: tk.StringVar
    max_files: int
    width: int
    height: int
    drop_frame: ctk.CTkFrame
    image_width: int = IMAGE_WIDTH
    columns: int
    rows: int

    @property
    def value(self) -> list[str]:
        return self.files

    @value.setter
    def value(self, value) -> None:
        pass

    @property
    def files(self) -> list[str]:
        return self.__value.copy()

    @files.setter
    def files(self, value: list[str]) -> None:
        if len(value) > self.max_files:
            raise TooManyFilesError(self.max_files)
        self.__value = value

    @property
    def message(self) -> str:
        return self.__msg.get()

    @message.setter
    def message(self, value: str) -> None:
        self.__msg.set(value)

    def __init__(
        self,
        label: str,
        *,
        max_files: int = 1,
        files: list[str] | None = None,
    ) -> None:
        self.label = label
        self.__value = files or []
        self.__msg = tk.StringVar(value="")
        self.max_files = max_files if max_files >= 1 else 1
        self.columns = MAX_COLUMNS if self.max_files > MAX_COLUMNS else self.max_files
        self.rows = (self.max_files - 1) // MAX_COLUMNS + 1
        self.width = (
            self.columns * self.image_width
            if self.max_files < 4
            else 4 * self.image_width
        )
        self.height = self.rows * self.image_width

    def __call__(self, master: tk.Frame) -> Self:
        self.master = master
        self.frame = ctk.CTkFrame(
            self.master,
            height=self.height,
            fg_color=ctk.ThemeManager.theme["CTkFrame"]["fg_color"],
        )
        return self

    def get(self) -> list[str]:
        return self.files

    def build(self, method: Literal["pack"] = "pack", **kwargs) -> Self:
        match method:
            case _:
                build_method = self.pack
        self.frame.pack()
        build_method(
            ctk.CTkLabel(master=self.frame, text=self.label), side=tk.LEFT, **kwargs
        )

        kwargs.pop("expand", True)
        self.drop_frame = ctk.CTkFrame(
            self.frame,
            width=self.width,
            height=self.height,
        )
        build_method(
            self.drop_frame,
            propagate=False,
            side=tk.LEFT,
            fill=tk.BOTH,
            expand=False,
            **kwargs,
        )
        self.drop_frame.drop_target_register(DND_FILES)  # type: ignore # pylint: disable=no-member
        self.drop_frame.dnd_bind("<<Drop>>", lambda file: self.add_file(file.data.strip("{}")))  # type: ignore  # pylint: disable=no-member
        # Message
        build_method(
            ctk.CTkLabel(self.master, textvariable=self.__msg), **gui.pack_defaults
        )

        button_frame = ctk.CTkFrame(
            self.frame, fg_color=ctk.ThemeManager.theme["CTkFrame"]["fg_color"]
        )
        build_method(button_frame, side=tk.RIGHT, **kwargs)

        if platform.system() != "Darwin":
            ctk.CTkButton(
                button_frame,
                text="Browse",
                command=lambda: self.add_files(
                    flowmatic.file_dialog(
                        title="Select File",
                        max_files=self.max_files,
                        filetypes=[("All Files", "*.*")],
                    )
                ),
            ).pack(**kwargs)

        ctk.CTkButton(
            button_frame,
            text="Clear",
            command=self.clear_files,
        ).pack(**kwargs)

        return self

    def pack(self, element: tk.Widget, **kwargs) -> None:
        if kwargs.pop("propagate", None) is False:
            element.pack_propagate(False)
            element.grid_propagate(False)
        element.pack(**kwargs)

    def add_files(self, files: Iterable[str]) -> None:
        for file in files:
            self.add_file(file)

    def add_file(self, file: str) -> None:
        try:
            self.files += [file]
        except TooManyFilesError as exc:
            self.message = str(exc)
        else:
            self.message = ""
            file_frame = ctk.CTkFrame(
                self.drop_frame,
                width=self.image_width,
                height=self.image_width,
                fg_color=ctk.ThemeManager.theme["CTkFrame"]["top_fg_color"],
            )
            index = len(self.files) - 1
            column, row = self.get_position(index)
            file_frame.grid(column=column, row=row)
            file_frame.master.grid_columnconfigure(column, weight=1)

            your_image = ctk.CTkImage(
                light_image=Image.open(
                    os.path.join(Icons.path, Icons.get_icon(file.lower()))
                ),
                size=(72, 97),
            )
            ctk.CTkLabel(master=file_frame, image=your_image, text="").pack(expand=True)
            ctk.CTkLabel(file_frame, text=os.path.split(file)[-1]).pack(
                **gui.pack_defaults
            )

    def get_position(self, index: int) -> tuple[int, int]:
        column = index % self.columns
        row = index // self.columns
        return column, row

    def clear_files(self) -> None:
        self.files = []
        self.message = ""
        for widget in self.drop_frame.winfo_children():
            widget.destroy()
        # flowmatic.update_gui()



from enum import StrEnum


class Icons(StrEnum):
    CODE = "Code.png"
    FILE = "File.png"
    FOLDER = "Folder.png"
    IMAGE = "Image.png"
    PDF = "PDF.png"
    TEXT = "Text.png"
    TABLE = "Table.png"
    ZIP = "Zip.png"

    @classmethod
    def get_icon(cls, file: str) -> str:
        extension = file.split(".")[-1]
        match extension:
            case "py" | "html" | "css" | "js" | "json" | "xml" | "yaml" | "yml":
                return cls.CODE
            case "pdf":
                return cls.PDF
            case "zip" | "rar" | "7z" | "tar" | "gz" | "xz" | "bz2":
                return cls.ZIP
            case "png" | "jpg" | "jpeg" | "gif" | "svg" | "ico" | "bmp" | "webp":
                return cls.IMAGE
            case "txt" | "md" | "rst" | "doc" | "docx" | "odt" | "rtf" | "tex" | "wpd" | "wps" | "odt":  # pylint: disable=line-too-long
                return cls.TEXT
            case "csv" | "tsv" | "xls" | "xlsx" | "ods":
                return cls.TABLE
            case _:
                return cls.FILE

    @classmethod
    @property
    def path(cls) -> str:
        return "assets/icons"



"""Utility Base Classes.

This module contains utility base classes that are used throughout the
application.

"""


from .saved_class import SavedClass
from .settings import Settings
from .icons import Icons
from .exceptions import *



import tkinter as tk
from typing import Any, Self

import customtkinter as ctk

from flowmatic import gui


class CheckBox:
    master: tk.Frame
    frame: tk.Frame
    label: str
    __value: tk.BooleanVar
    show: bool

    @property
    def value(self) -> bool:
        return self.__value.get()

    @value.setter
    def value(self, value: bool) -> None:
        self.__value.set(value)

    def __init__(
        self,
        label: str,
        *,
        value: bool | None = None,
    ) -> None:
        self.label = label
        self.__value = tk.BooleanVar(value=value)

    def __call__(self, master: tk.Frame) -> Self:
        self.master = master
        self.frame = ctk.CTkFrame(
            self.master,
            fg_color=ctk.ThemeManager.theme["CTkFrame"]["fg_color"],
        )
        return self

    def build(self, **kwargs: dict[str, Any]) -> Self:
        self.frame.pack(**gui.pack_defaults)
        ctk.CTkLabel(master=self.frame, text=self.label).pack(side=tk.LEFT, **kwargs)
        ctk.CTkCheckBox(self.frame, variable=self.__value).pack(side="left", **kwargs)
        return self

    def pack(self, element: tk.Widget, **kwargs) -> None:
        element.pack(**kwargs)



from .file_field import FileField
from .radio_group import RadioGroup
from .text_field import TextField
from .field import Field
from .display import Display
from .check_box import CheckBox
from .date_select import DateSelect



from typing import Callable, NamedTuple


Element = NamedTuple


class Button(Element):
    text: str
    command: Callable[[], None]


class Info(Element):
    label: str
    value: str


class ElementFrame(Element):
    title: str
    rows: list[Element]



from .form_screen import FormScreen, FlowFormScreen
from .menu_screen import MenuScreen
from .title_screen import TitleScreen
from .screen import FlowScreen, Screen, WrappedScreen
from .element_screen import ElementScreen
from .result_screen import ResultScreen, Result, Status



import tkinter as tk
from typing import Callable, TypeVar, overload

import customtkinter as ctk

import flowmatic
from flowmatic import gui
from flowmatic.flows.flow import Flow
from flowmatic.gui.screens.screen import FlowScreen, Screen
from ...forms import FormValidation
from ...forms.fields.field import Field
from ..elements.element_infos import Button

T = TypeVar("T")


class FormScreen(Screen):
    field_list: list[Field] | None
    fields: dict[str, Field]
    validations: list[FormValidation] | None = None
    __message: tk.StringVar

    @property
    def message(self) -> str:
        return self.__message.get()

    @message.setter
    def message(self, value: str) -> None:
        """Set message.

        Args:
            value (str): Message."""
        self.__message.set(value)

    def __init__(
        self,
    ) -> None:
        self.fields = {}
        self.__message = tk.StringVar(value="")

    @property
    def values(self) -> dict[str, str | int | list[str]]:
        """Get values from fields.

        Returns:
            dict[str, str]: Values from fields."""
        return {
            label: field.value for label, field in self.fields.items() if field.value
        }

    def validate(self) -> None:
        if not self.validations:
            return

        for validation in self.validations:
            validation.validate(self.values)

    @overload
    def get(self, __key: str, /) -> str | int | list[str]:
        ...

    @overload
    def get(self, __key: str, __default: T) -> str | int | list[str] | T:
        ...

    @overload
    def get(self, __key: str, __default: T, __return: type[T]) -> T:
        ...

    def get(
        self, __key: str, __default: str | T = "", __return: type[T] | None = None
    ) -> str | int | list[str] | T:
        """Get value from field.

        Args:
            label (str): Label of field.

        Returns:
            str: Value of field."""
        value = self.values.get(__key, __default)
        if __return is not None:
            if not isinstance(value, __return):
                raise TypeError(f"Expected {__return} got {type(value)}")
        return value

    def submit_default(self) -> None:
        flowmatic.show_start_screen

    def build(  # pylint: disable=arguments-differ
        self,
        *,
        title: str,
        validations: list[FormValidation] | None = None,
        submit_button: Button | None = None,
        back_button: Button | None = None,
        fields: list[Field] | None = None,
        field_factory: Callable[..., list[Field]] | None = None,
    ) -> None:
        """Build screen.

        Args:

            title (str): Title of screen.
            validate_command (Callable[[], None]): Command to run when validate button is pressed.
            validate_text (str, optional): Text of validate button. Defaults to "Submit".
            back_command (Callable[[], None], optional): Command to run when back button is pressed.
                 Defaults to None.
        """
        # Variables
        self.field_list = (fields or []) + (field_factory() if field_factory else [])
        self.validations = validations
        if not submit_button:
            submit_button = Button(text="Submit", command=self.submit_default)

        def validate_command() -> None:
            try:
                self.validate()
            except AssertionError as error:
                self.message = str(error)
                flowmatic.update_gui()
            else:
                submit_button.command()

        # Title
        ctk.CTkLabel(self.master, text=title).pack(**gui.pack_defaults)

        # Fields
        if self.field_list:
            fields_frame = (
                ctk.CTkScrollableFrame(
                    self.master, height=gui.HEIGHT - 150, width=gui.WIDTH
                )
                if not isinstance(self.master, ctk.CTkScrollableFrame)
                else ctk.CTkFrame(self.master)
            )
            fields_frame.pack(**gui.pack_defaults)
            for field in self.field_list:
                self.fields[field.label] = field(fields_frame).build(
                    **gui.pack_defaults
                )

        # Message
        ctk.CTkLabel(self.master, textvariable=self.__message).pack(**gui.pack_defaults)

        # Buttons
        pack_args = gui.pack_defaults  # | {"pady": (10, 20)}
        button_frame = ctk.CTkFrame(self.master)
        button_frame.pack(**pack_args)
        ctk.CTkButton(
            button_frame, text=submit_button.text or "Submit", command=validate_command
        ).pack(**gui.pack_defaults, side=tk.RIGHT)
        if back_button:
            ctk.CTkButton(
                button_frame,
                text=back_button.text or "Back",
                command=back_button.command,
            ).pack(**gui.pack_defaults, side=tk.LEFT)


class FlowFormScreen(FormScreen, FlowScreen):
    def __init__(self, flow: Flow) -> None:
        self.flow = flow
        super().__init__()



from typing import Protocol, Self
import tkinter as tk
import customtkinter as ctk

from ...flows.flow import Flow
from flowmatic import gui

TOPBAR_HEIGHT = 100
SIDEBAR_WIDTH = 200


class Screen(Protocol):
    """Screen protocol.

    Args:
        master (tk.Tk): Master window.

    Methods:
        build: Build screen."""

    master: tk.Tk | tk.Frame

    def __init__(self) -> None:
        ...

    def __call__(self, master: tk.Tk | tk.Frame) -> Self:
        self.master = master  # pylint: disable=attribute-defined-outside-init
        return self

    def build(self) -> None:  # pylint: disable=missing-function-docstring
        ...


class FlowScreen(Screen):
    flow: Flow

    def __init__(self, flow: Flow) -> None:
        self.flow = flow
        super().__init__()

    def next(self) -> None:
        self.flow.next()

    def previous(self) -> None:
        self.flow.previous()

    def build(self) -> None:
        raise NotImplementedError


class WrappedScreen(Screen):
    screens: list[Screen]
    inner: type[Screen]
    sidebar: type[Screen] | None = None
    topbar: type[Screen] | None = None

    def __init__(self) -> None:
        self.screens = []

    def __call__(self, master: tk.Tk) -> Self:
        self.master = master
        outer_frame = ctk.CTkFrame(master)
        outer_frame.pack(fill=tk.BOTH, expand=True)
        inner_frame = ctk.CTkScrollableFrame(
            outer_frame,
            width=gui.WIDTH - SIDEBAR_WIDTH,
            height=gui.HEIGHT - TOPBAR_HEIGHT,
        )
        self.screens.append(self.inner()(inner_frame))
        if self.sidebar is not None:
            side_frame = ctk.CTkFrame(outer_frame, width=SIDEBAR_WIDTH)
            side_frame.pack(side=tk.LEFT)
            self.screens.append(self.sidebar()(side_frame))
        if self.topbar is not None:
            top_frame = ctk.CTkFrame(outer_frame, height=TOPBAR_HEIGHT)
            top_frame.pack(side=tk.TOP)
            self.screens.append(self.topbar()(top_frame))

        inner_frame.pack()
        return self

    def build(self) -> None:
        for screen in self.screens:
            screen.build()



from typing import Protocol
from .screens.screen import Screen
from ..flows.flow import Flow


class GUI(Protocol):
    title: str
    start_screen: Screen
    current_screen: Screen
    current_flow: Flow

    def __init__(self, title: str, start_screen: Screen):
        ...

    def start(self, start_screen: Screen | None):
        ...

    def setup(self):
        ...

    def switch_screen(self, screen: Screen):
        ...

    def start_flow(self, flow: Flow):
        ...

    def build_screen(self, screen: Screen):
        ...

    def clear_screen(self):
        ...

    def file_dialog(
        self, title: str, max_files: int, filetypes: list[tuple[str, str]]
    ) -> str | tuple[str, ...]:
        ...

    def show_error(self, error: Exception | str) -> None:
        ...

    def update(self):
        ...

    def quit(self):
        ...



import datetime
import tkinter as tk
from typing import Self
import tkcalendar as tkc
import customtkinter as ctk
from flowmatic import gui


class DateSelect:
    master: tk.Frame
    frame: tk.Frame
    label: str
    __value: tk.StringVar

    @property
    def value(self) -> str:
        return self.__value.get()

    @value.setter
    def value(self, value: str) -> None:
        self.__value.set(value)

    def __init__(
        self,
        label: str,
        *,
        value: str | None = None,
    ) -> None:
        self.__value = tk.StringVar(
            value=value or datetime.date.today().strftime("%d.%m.%Y")
        )
        self.label = label

    def __call__(self, master: tk.Frame) -> Self:
        self.master = master
        self.frame = ctk.CTkFrame(
            self.master,
            fg_color=ctk.ThemeManager.theme["CTkFrame"]["fg_color"],
        )
        return self

    def build(self, **kwargs) -> Self:
        self.frame.pack()
        ctk.CTkLabel(master=self.frame, text=self.label).pack(
            side=tk.LEFT, **gui.pack_defaults
        )
        calendar = tkc.Calendar(
            self.frame,
            selectmode="day",
            date_pattern="dd.mm.yyyy",
            showweeknumbers=True,
            textvariable=self.__value,
        )
        calendar.pack(side=tk.RIGHT, **gui.pack_defaults)
        return self

    def pack(self, element: tk.Widget, **kwargs) -> None:
        element.pack(**kwargs)



"""Exceptions used by Appy.

Classes:
    SavedClassLoadException: Raised when a class cannot be loaded from a saved file."""


class SavedClassLoadException(Exception):
    path: str

    def __init__(self, path: str) -> None:
        self.path = path

    def __str__(self) -> str:
        return f"Could not load class from {self.path}"


class TooManyFilesError(Exception):
    max_files: int

    def __init__(self, max_files: int) -> None:
        self.max_files = max_files

    def __str__(self) -> str:
        return f"Too many files. Max: {self.max_files}"


class NotAFileError(Exception):
    path: str

    def __init__(self, path: str) -> None:
        self.path = path

    def __str__(self) -> str:
        return f"Not a file: {self.path}"



from .gui import GUI, TKGUI
from .gui.screens.screen import Screen
from .util.settings import Settings
from .util.user import User
from .flows.flow import Flow


class App:
    """Singleton for accessing GUI, settings, user and other app-wide objects.

    Attributes:
        gui (GUI): GUI interface.
        settings (Settings): Settings.
        user (User): User.

    Args:
        title (str, optional): Title of the app. Defaults to "Appy".
    """

    title: str = "FlowMatic"
    gui: GUI = None  # type: ignore
    settings: Settings
    user: User

    start_screen: Screen

    def __init__(self, start_screen: type[Screen]):
        self.gui = self.gui or TKGUI(
            title=self.title, start_screen=start_screen, geometry="1280x720"
        )
        self.start_screen = start_screen()
        self.settings = Settings.load("files/settings.json")
        self.user = User("")

    def start(self, start_screen: Screen | None = None):
        """Start app.
        Args:
            screen (type[Screen]): Screen to start with."""
        self.gui.start(start_screen)

    def show_start_screen(self):
        """Show start screen."""
        self.gui.switch_screen(self.start_screen)

    def show_screen(self, screen: Screen):
        """Show start screen.
        Args:
            screen (type[Screen]): Screen to show."""
        self.gui.switch_screen(screen)

    def start_flow(self, flow: Flow):
        """Start flow.
        Args:
            flow (type[Flow]): Flow to start."""
        self.gui.start_flow(flow)

    def file_dialog(
        self, title: str, max_files: int, filetypes: list[tuple[str, str]]
    ) -> tuple[str, ...]:
        """Open file dialog.
        Args:
            title (str): Title of the dialog.
            max_files (int): Maximum number of files to select.
            filetypes (list[tuple[str, str]]): Filetypes to show.
        Returns:
            str | None: Path to file or None if no file selected."""
        filenames = self.gui.file_dialog(title, max_files, filetypes)

        if not filenames:
            return ("",)

        if isinstance(filenames, str):
            return (filenames,)

        return filenames

    def quit(self):
        self.gui.quit()



from __future__ import annotations
from typing import TYPE_CHECKING, Protocol

import flowmatic

if TYPE_CHECKING:
    from flowmatic.gui.screens.screen import FlowScreen, Screen


class Flow(Protocol):
    steps: list[type[FlowScreen]]
    screens: list[FlowScreen]
    return_to: Screen
    step: int

    def __init__(self) -> None:
        self.screens = [screen(self) for screen in self.steps]
        self.step = 0

    def __next__(self) -> None:
        if self.step == len(self.screens) or self.step < 0:
            flowmatic.show_screen(self.return_to)
            return
        screen = self.screens[self.step]
        self.step += 1
        flowmatic.show_screen(screen)

    def start(self):
        next(self)

    def quit(self):
        self.step = len(self.screens)
        next(self)

    def next(self):
        next(self)

    def previous(self):
        self.step -= 2
        next(self)



import tkinter as tk
from typing import Any, Self

import customtkinter as ctk


class TextField:
    master: tk.Frame
    frame: tk.Frame
    label: str
    __value: tk.StringVar
    placeholder_text: str | None
    show: bool

    @property
    def value(self) -> str:
        return self.__value.get()

    @value.setter
    def value(self, value: str) -> None:
        self.__value.set(value)

    def __init__(
        self,
        label: str,
        *,
        value: str | None = None,
        placeholder_text: str | None = None,
        show: bool = True
    ) -> None:
        self.label = label
        self.__value = tk.StringVar(value=value)
        self.placeholder_text = placeholder_text
        self.show = show

    def __call__(self, master: tk.Frame) -> Self:
        self.master = master
        self.frame = ctk.CTkFrame(
            self.master,
            fg_color=ctk.ThemeManager.theme["CTkFrame"]["fg_color"],
        )
        return self

    def __eq__(self, __value: object) -> bool:
        if not isinstance(__value, TextField):
            return False
        return self.value == __value.value and self.label == __value.label

    def build(self, **kwargs: dict[str, Any]) -> Self:
        self.frame.pack()
        ctk.CTkLabel(master=self.frame, text=self.label).pack(side=tk.LEFT, **kwargs)
        entry = ctk.CTkEntry(
            self.frame,
            textvariable=self.__value,
            placeholder_text=self.placeholder_text,
            show="•" if not self.show else None,
        )
        entry.pack(side=tk.RIGHT, **kwargs)
        return self

    def pack(self, element: tk.Widget, **kwargs) -> None:
        element.pack(**kwargs)



import customtkinter as ctk

from flowmatic import gui
from flowmatic.gui.screens.screen import Screen
from ...gui.elements import Button


class MenuScreen(Screen):
    """Menu screen. Used for menus.

    Args:
        master (tk.Tk): Master window."""

    def build(  # pylint: disable=arguments-differ
        self, title: str, buttons: list[Button]
    ) -> None:
        """Build screen.

        Args:
            title (str): Title of screen.
            buttons (list[gui.ButtonInfo]): List of buttons to add to screen."""
        ctk.CTkLabel(self.master, text=title).pack(**gui.pack_defaults)
        for button in buttons:
            ctk.CTkButton(self.master, text=button.text, command=button.command).pack(
                **gui.pack_defaults
            )



import customtkinter as ctk
import tkinter as tk
from flowmatic import gui
from flowmatic.gui.elements.element_infos import *  # pylint: disable=wildcard-import, unused-wildcard-import
from flowmatic.gui.screens.screen import Screen


class ElementScreen(Screen):
    def build(  # pylint: disable=arguments-differ
        self, elements: list[Element], **kwargs
    ) -> None:
        for element in elements:
            self.build_element(element, self.master, **kwargs)

    @classmethod
    def build_element(
        cls, element: Element, master: tk.Frame | tk.Tk, **kwargs
    ) -> None:
        match element:
            case Button():
                ctk.CTkButton(master, text=element.text, command=element.command).pack(
                    **(gui.pack_defaults | kwargs),
                )
            case Info():
                ctk.CTkLabel(
                    master,
                    text=f"{element.label}: {element.value}",
                ).pack(**(gui.pack_defaults | kwargs))
            case ElementFrame():
                frame = ctk.CTkFrame(master)
                frame.pack(**gui.pack_defaults)
                ctk.CTkLabel(frame, text=element.title).pack(
                    **(gui.pack_defaults | kwargs),
                )
                for row in element.rows:
                    cls.build_element(row, frame, **kwargs)
            case _:
                raise NotImplementedError(
                    f"Unknown element type: {element.__class__.__name__}, extend ElementScreen.build_element() to support it."
                )
